/*
 * Quoridor server and related libraries
 * Copyright (C) 2009-2010 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://www.gnu.org/licenses/.
 */
package cz.xelfi.quoridor.freemarkerdor;

import java.util.Collection;
import java.util.Arrays;
import com.sun.jersey.api.client.Client;
import com.sun.jersey.api.client.ClientResponse;
import com.sun.jersey.api.client.WebResource;
import com.sun.jersey.core.header.MediaTypes;
import com.sun.jersey.core.util.MultivaluedMapImpl;
import com.sun.net.httpserver.HttpServer;
import cz.xelfi.quoridor.webidor.resources.Quoridor;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.HashSet;
import java.util.Locale;
import java.util.concurrent.Callable;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.ws.rs.core.Cookie;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.MultivaluedMap;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
public class UITest extends Object {
    private static File dir;
    private static HttpServer stopAPI;
    private static HttpServer stopStatistics;
    private static Callable<Void> stop;

    private WebResource webResource;
    private WebResource apiResource;
    private WebResource statResource;

    public UITest() throws Exception {
    }

    @BeforeClass
    public static void localeEnglish() throws Exception {
        Locale.setDefault(Locale.ENGLISH);
        dir = File.createTempFile("quoridor", ".dir");
        dir.delete();
        dir.mkdirs();
        
        File games = new File(dir, "games");
        copyResource("wrong.encoding", new File(games, "fdcc275c-55ad-4b7e-a776-bde3876b2f2c"));
        
        
        System.setProperty("quoridor.dir", dir.getPath());
        stopAPI = Quoridor.start(9990);
        stopStatistics = Quoridor.start(9992);
        stop = UI.startServers(9991, "http://localhost:9990", "http://localhost:9992", null);

        File passwd = new File(dir, "passwd");
        FileOutputStream os = new FileOutputStream(passwd);
        os.write("test=pes\nJarda=darda\n".getBytes("UTF-8"));
        os.close();
    }

    @Before
    public void setUp() throws Exception {

        Client client = new Client();
        webResource = client.resource(new URI("http://localhost:9991/"));
        apiResource = client.resource(new URI("http://localhost:9990/"));
        statResource = client.resource(new URI("http://localhost:9992/"));
    }

    @AfterClass
    public static void cleanUpAll() throws Exception {
        deleteRec(dir);
        if (stop != null) {
            stop.call();
        }
        if (stopAPI != null) {
            stopAPI.stop(0);
        }
    }

    static void deleteRec(File dir) throws IOException {
        if (dir == null) {
            return;
        }
        File[] arr = dir.listFiles();
        if (arr != null) {
            for (File f : arr) {
                deleteRec(f);
            }
        }
        dir.delete();
    }

    @Test public void testApplicationWadl() {
        String serviceWadl = webResource.path("application.wadl").
                accept(MediaTypes.WADL).get(String.class);
        assertTrue(serviceWadl.length() > 0);
    }

    @Test public void testAPIWadl() {
        String serviceWadl = apiResource.path("application.wadl").
                accept(MediaTypes.WADL).get(String.class);
        assertTrue(serviceWadl.length() > 0);
    }

    @Test public void testStatWadl() {
        String serviceWadl = statResource.path("application.wadl").
                accept(MediaTypes.WADL).get(String.class);
        assertTrue(serviceWadl.length() > 0);
    }

    @Test public void testGetIndexPage() throws Exception {
        String logJarda = apiResource.path("login").
            queryParam("name", "Jarda").
            queryParam("password", "darda").
            accept(MediaType.TEXT_PLAIN).
            put(String.class);
        assertNotNull("Logged in OK", logJarda);

        String res = webResource.accept("text/html").get(String.class);
        if (res.indexOf("Quoridor") == -1) {
            fail("Wrong index.html:\n" + res);
        }
        if (res.indexOf("Login") == -1) {
            fail("Wrong index.html:\n" + res);
        }
        if (res.indexOf("action=\"/login\"") == -1) {
            fail("Wrong index.html:\n" + res);
        }
        if (res.toLowerCase().indexOf("error") != -1) {
            fail("There was an error:\n" + res);
        }
        res = webResource.cookie(Cookie.valueOf("login=" + logJarda)).accept("text/html").get(String.class);
        if (res.indexOf("action=\"/games/create\"") == -1) {
            fail(res);
        }
    }

    @Test public void testLogin() throws Exception {
        MultivaluedMap formData = new MultivaluedMapImpl();
        formData.add("name", "test");
        formData.add("password", "pes");
        ClientResponse response = webResource.path("login").
            accept("text/html").type("application/x-www-form-urlencoded").
            post(ClientResponse.class, formData);
        final String res = response.getEntity(String.class);
        assertEquals("OK", ClientResponse.Status.OK, response.getClientResponseStatus());
        if (res.indexOf("You are logged in as test") == -1) {
            fail("res: " + res);
        }
    }


    @Test public void testCreateGameWrongUsers() throws Exception {
        String logTest = apiResource.path("login").
            queryParam("name", "test").
            queryParam("password", "pes").
            accept(MediaType.TEXT_PLAIN).
            put(String.class);
        ClientResponse res = webResource.path("games/create").
            queryParam("white", "unknown1").
            queryParam("black", "unknown2").
            cookie(Cookie.valueOf("login=" + logTest)).
            accept("text/html").
            get(ClientResponse.class);
        final String txt = res.getEntity(String.class);
        assertEquals("Rejected, unknown user\n" + txt, ClientResponse.Status.NOT_FOUND, res.getClientResponseStatus());
        if (txt.indexOf("You (test) must be") == -1) {
            fail(txt);
        }
    }

    @Test public void testCreateGameOK() throws Exception {
        Collection<String> origGamesNames= Arrays.asList(new File(dir, "games").list());
        int origGames = origGamesNames.size();
        
        String logTest = apiResource.path("login").
            queryParam("name", "test").
            queryParam("password", "pes").
            accept(MediaType.TEXT_PLAIN).
            put(String.class);
        String logJarda = apiResource.path("login").
            queryParam("name", "Jarda").
            queryParam("password", "darda").
            accept(MediaType.TEXT_PLAIN).
            put(String.class);
        ClientResponse res = webResource.path("games/create").
            queryParam("white", "Jarda").
            queryParam("black", "test").
            cookie(Cookie.valueOf("login=" + logTest)).
            accept("text/html").
            get(ClientResponse.class);

        final String txt = res.getEntity(String.class);
        assertEquals("OK\n" + txt, ClientResponse.Status.OK, res.getClientResponseStatus());

        
        String myGame;
        {
            String[] games = new File(dir, "games").list();
            assertEquals("One new game", origGames + 1, games.length);
            
            HashSet<String> newNames = new HashSet<String>();
            newNames.addAll(Arrays.asList(games));
            newNames.removeAll(origGamesNames);
            assertEquals("Just one name left: " + newNames, 1, newNames.size());
            myGame = newNames.iterator().next();
        }

        if (txt.indexOf(myGame) == -1) {
            fail(myGame + " expected inside of:\n" + txt);
        }

        ClientResponse page = webResource.path("games/" + myGame).
            cookie(Cookie.valueOf("login=" + logJarda)).
            get(ClientResponse.class);
        String ptxt = page.getEntity(String.class);
        assertEquals("OK:\n" + ptxt, ClientResponse.Status.OK, page.getClientResponseStatus());

        Pattern p = Pattern.compile(".*<img[^>]*src=\"([^\"]*)\"");
        Matcher m = p.matcher(ptxt);
        assertTrue("image found\n" + ptxt, m.find());

        InputStream img1 = webResource.path(m.group(1)).get(InputStream.class);
        assertNotNull("image found", img1);

        ClientResponse move = apiResource.path("games/" + myGame).
            queryParam("loginID", logJarda).
            queryParam("player", "Jarda").queryParam("move", "N").put(ClientResponse.class);
        assertEquals("Move OK:\n" + move.getEntity(String.class), ClientResponse.Status.OK, move.getClientResponseStatus());

        InputStream img2 = webResource.path(m.group(1)).get(InputStream.class);
        assertNotNull("image found", img2);

        ClientResponse page2 = webResource.path("games/" + myGame).
            cookie(Cookie.valueOf("login=" + logJarda)).
            get(ClientResponse.class);
        String ptxt2 = page2.getEntity(String.class);
        assertEquals("OK:\n" + ptxt2, ClientResponse.Status.OK, page2.getClientResponseStatus());

        Matcher m2 = p.matcher(ptxt2);
        assertTrue("image found\n" + ptxt2, m2.find());

        InputStream img3 = webResource.path(m2.group(1)).get(InputStream.class);
        assertNotNull("image found", img3);

        int diff = 0;
        int cnt = 0;
        for (;;) {
            cnt++;
            int b1 = img1.read();
            int b2 = img2.read();
            int b3 = img3.read();
            assertEquals(b1, b2);
            if (b3 != b1) {
                diff++;
            }
            if (b1 == -1 || b3 == -1) break;

        }
        if (diff == 0) {
//            fail("There shall be difference in the streams. Read bytes " + cnt);
        }
    }
    
    @Test public void testGetWrongGame() throws Exception {
        String logTest = apiResource.path("login").
            queryParam("name", "test").
            queryParam("password", "pes").
            accept(MediaType.TEXT_PLAIN).
            put(String.class);
        
        String xml = apiResource.path("games/fdcc275c-55ad-4b7e-a776-bde3876b2f2c").
            queryParam("name", "test").
            queryParam("password", "pes").
            accept(MediaType.TEXT_XML).
            get(String.class);
        
        if (xml.indexOf((char)0x1b) >= 0) {
            fail(xml.replace((char)0x1b, 'X'));
        }
        
        ClientResponse res = webResource.path("games/fdcc275c-55ad-4b7e-a776-bde3876b2f2c").
            queryParam("white", "Jarda").
            queryParam("black", "test").
            cookie(Cookie.valueOf("login=" + logTest)).
            accept("text/html").
            get(ClientResponse.class);
        final String txt = res.getEntity(String.class);
        assertEquals("OK\n" + txt, ClientResponse.Status.OK, res.getClientResponseStatus());
    }
    

    private static void copyResource(String res, File file) throws IOException {
        InputStream is = UITest.class.getResourceAsStream(res);
        file.getParentFile().mkdirs();
        FileOutputStream os = new FileOutputStream(file);
        byte[] arr = new byte[4092];
        for (;;) {
            int len = is.read(arr);
            if (len == -1) {
                break;
            }
            os.write(arr, 0, len);
        }
        is.close();
        os.close();
    }
    
}
